@@ -26,6 +26,7 @@
 #include <sys/types.h>
 #include <netinet/in.h>
 #include <sys/socket.h>
+#include <sys/un.h>
 #include <arpa/inet.h>
 #include <sys/select.h>
 #include <netinet/in.h>
@@ -320,9 +321,9 @@ MsgChannel::operator>> (uint32_t &buf)
   if (inofs >= intogo + 4)
     {
       if ( ptrdiff_t(inbuf + intogo) % 4 ) {
-        char t_buf[4];
+        uint32_t t_buf[1];
         memcpy(t_buf, inbuf + intogo, 4);
-        buf = *(uint32_t *)t_buf;
+        buf = t_buf[0];
       } else
         buf = *(uint32_t *)(inbuf + intogo);
       intogo += 4;
@@ -656,10 +657,35 @@ MsgChannel *Service::createChannel (const string &hostname, unsigned short p, in
       if (connect (remote_fd, (struct sockaddr *) &remote_addr, sizeof (remote_addr)) < 0)
         {
           close( remote_fd );
-	  trace() << "connect failed\n";
+	  trace() << "connect failed on " << hostname << endl;
           return 0;
         }
     }
+  trace() << "connected to " << hostname << endl;
+  return createChannel(remote_fd, (struct sockaddr *)&remote_addr, sizeof( remote_addr ));
+}
+
+MsgChannel *Service::createChannel (const string &socket_path)
+{
+  int remote_fd;
+  struct sockaddr_un remote_addr;
+
+  if ((remote_fd = socket (AF_UNIX, SOCK_STREAM, 0)) < 0)
+    {
+      log_perror("socket()");
+      return 0;
+    }
+
+  remote_addr.sun_family = AF_UNIX;
+  strncpy(remote_addr.sun_path, socket_path.c_str(), sizeof(remote_addr.sun_path) - 1);
+
+  if (connect (remote_fd, (struct sockaddr *) &remote_addr, sizeof (remote_addr)) < 0)
+    {
+      close( remote_fd );
+      trace() << "connect failed on " << socket_path << endl; 
+      return 0;
+    }
+  trace() << "connected to " << socket_path << endl;
   return createChannel(remote_fd, (struct sockaddr *)&remote_addr, sizeof( remote_addr ));
 }
 
@@ -850,11 +876,16 @@ MsgChannel::wait_for_msg (int timeout)
 {
   if (has_msg ())
     return true;
-  if (!read_a_bit () || timeout <= 0)
+  if (!read_a_bit ())
     {
-      trace() << "!read_a_bit || timeout <= 0\n";
+      trace() << "!read_a_bit\n";
       return false;
     }
+  if (timeout <= 0)
+    {
+      trace() << "timeout <= 0\n";
+      return has_msg ();
+    }
   while (!has_msg ())
     {
       fd_set read_set;
@@ -884,8 +915,8 @@ MsgChannel::get_msg(int timeout)
   Msg *m = 0;
   enum MsgType type;
   uint32_t t;
-  if (timeout > 0 && !wait_for_msg (timeout) ) {
-    trace() << "blocking && !waiting_for_msg()\n";
+  if (!wait_for_msg (timeout) ) {
+    trace() << "!wait_for_msg()\n";
     return 0;
   }
   /* If we've seen the EOF, and we don't have a complete message,
@@ -935,6 +966,9 @@ MsgChannel::get_msg(int timeout)
     case M_GET_INTERNALS: m = new GetInternalStatus; break;
     case M_STATUS_TEXT: m = new StatusTextMsg; break;
     case M_CS_CONF: m = new ConfCSMsg; break;
+    case M_VERIFY_ENV: m = new VerifyEnvMsg; break;
+    case M_VERIFY_ENV_RESULT: m = new VerifyEnvResultMsg; break;
+    case M_BLACKLIST_HOST_ENV: m = new BlacklistHostEnvMsg; break;
     case M_TIMEOUT: break;
     }
   if (!m) {
@@ -987,6 +1021,12 @@ open_send_broadcast (void)
       return -1;
     }
 
+  if (fcntl (ask_fd, F_SETFD, FD_CLOEXEC) < 0)
+    {
+      log_perror("open_send_broadcast fcntl");
+      close (ask_fd);
+      return -1;
+    }
   int optval = 1;
   if (setsockopt (ask_fd, SOL_SOCKET, SO_BROADCAST, &optval, sizeof(optval)) < 0)
     {
@@ -1229,6 +1269,14 @@ GetCSMsg::fill_from_channel (MsgChannel *c)
   preferred_host = string();
   if (IS_PROTOCOL_22(c))
     *c >> preferred_host;
+  if (IS_PROTOCOL_31(c))
+    {
+      uint32_t ign;
+      *c >> ign;
+      ignore_unverified = ( ign != 0 );
+    }
+  else
+    ignore_unverified = false;
 }
 
 void
@@ -1244,6 +1292,8 @@ GetCSMsg::send_to_channel (MsgChannel *c) const
   *c << client_id;
   if (IS_PROTOCOL_22(c))
     *c << preferred_host;
+  if (IS_PROTOCOL_31(c))
+    *c << uint32_t( ignore_unverified );
 }
 
 void
@@ -1301,6 +1351,12 @@ CompileFileMsg::fill_from_channel (MsgChannel *c)
   string target;
   *c >> target;
   job->setTargetPlatform( target );
+  if (IS_PROTOCOL_30(c))
+  {
+    string compilerName;
+    *c >> compilerName;
+    job->setCompilerName( compilerName );
+  }
 }
 
 void
@@ -1309,10 +1365,35 @@ CompileFileMsg::send_to_channel (MsgChannel *c) const
   Msg::send_to_channel (c);
   *c << (uint32_t) job->language();
   *c << job->jobID();
-  *c << job->remoteFlags();
+  if (IS_PROTOCOL_30(c))
+    *c << job->remoteFlags();
+  else
+  {
+    if (job->compilerName().find("clang") != string::npos)
+    { // Hack for compilerwrapper.
+        std::list<std::string> flags = job->remoteFlags();
+        flags.push_front("clang");
+        *c << flags;
+    }
+    else
+      *c << job->remoteFlags();
+  }
   *c << job->restFlags();
   *c << job->environmentVersion();
   *c << job->targetPlatform();
+  if (IS_PROTOCOL_30(c))
+    *c << remote_compiler_name();
+}
+
+// Environments created by icecc-create-env always use the same binary name
+// for compilers, so even if local name was e.g. c++, remote needs to
+// be g++ (before protocol version 30 remote CS even had /usr/bin/{gcc|g++}
+// hardcoded).  For clang, the binary is just clang for both C/C++.
+string CompileFileMsg::remote_compiler_name() const
+{
+    if (job->compilerName().find("clang") != string::npos)
+        return "clang";
+    return job->language() == CompileJob::Lang_CXX ? "g++" : "gcc";
 }
 
 CompileJob *
@@ -1542,6 +1623,26 @@ StatsMsg::send_to_channel (MsgChannel *c) const
   *c << freeMem;
 }
 
+void
+GetNativeEnvMsg::fill_from_channel (MsgChannel *c)
+{
+  Msg::fill_from_channel (c);
+  if (IS_PROTOCOL_32(c)) {
+    *c >> compiler;
+    *c >> extrafiles;
+  }
+}
+
+void
+GetNativeEnvMsg::send_to_channel (MsgChannel *c) const
+{
+  Msg::send_to_channel (c);
+  if (IS_PROTOCOL_32(c)) {
+    *c << compiler;
+    *c << extrafiles;
+  }
+}
+
 void
 UseNativeEnvMsg::fill_from_channel (MsgChannel *c)
 {
@@ -1682,6 +1783,56 @@ StatusTextMsg::send_to_channel (MsgChannel *c) const
   *c << text;
 }
 
+void
+VerifyEnvMsg::fill_from_channel (MsgChannel *c)
+{
+  Msg::fill_from_channel( c );
+  *c >> environment;
+  *c >> target;
+}
+
+void
+VerifyEnvMsg::send_to_channel (MsgChannel *c) const
+{
+  Msg::send_to_channel( c );
+  *c << environment;
+  *c << target;
+}
+
+void
+VerifyEnvResultMsg::fill_from_channel (MsgChannel *c)
+{
+  Msg::fill_from_channel( c );
+  uint32_t read_ok;
+  *c >> read_ok;
+  ok = read_ok != 0;
+}
+
+void
+VerifyEnvResultMsg::send_to_channel (MsgChannel *c) const
+{
+  Msg::send_to_channel( c );
+  *c << uint32_t( ok );
+}
+
+void
+BlacklistHostEnvMsg::fill_from_channel (MsgChannel *c)
+{
+  Msg::fill_from_channel( c );
+  *c >> environment;
+  *c >> target;
+  *c >> hostname;
+}
+
+void
+BlacklistHostEnvMsg::send_to_channel (MsgChannel *c) const
+{
+  Msg::send_to_channel( c );
+  *c << environment;
+  *c << target;
+  *c << hostname;
+}
+
 /*
 vim:cinoptions={.5s,g0,p5,t0,(0,^-0.5s,n-0.5s:tw=78:cindent:sw=4:
 */
